/***************************************************************************
 *   Copyright (C) 2007 by E. Angelou, A. Sottoriva                        *
 *   Universiteit van Amsterdam                                            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef ADVANCED_H_
#define ADVANCED_H_

#include "basic.h"

//R matrix updating and Theta sampling
int dustmat(double& p1, double& p2, double& p3, double& p4, double&	cost, double& cost2, double& pl, double& pc, double& sc, double& hgg, double& g2, double& pi) {


    double cosf, cosf2;
    double tmp;
    double f, f2, theta;

    //Read the accompanying paper for an explanation of the dust
    //scattering matrix (Dust radiative transfer using Monte Carlo methods, 
    //E.Angelou, A.Sottoriva, 2007)
    
    //Compute p1...p3 from (21),(22),(23)
    tmp = (double) (g2 + 1 - hgg * 2 * cost);
    p1 = (1 - g2) / pow(tmp, 1.5);
    p2 = -(pl) * p1 * (1 - cost2) / (cost2 + 1);
    p3 = p1 * 2 * cost / (cost2 + 1);


    if (fabs(cost) > 1.0) {
        if (cost > 1.0) {
            cost = 1.0;
        } else {
            cost = -1.0;
        }
    }

    //Compute Theta_f (f2) 
    //get Theta in degrees
    theta = acos(cost) * 180.0 / pi;
    //some calculations in degrees
    f = theta * 3.13f * exp(theta * -7.0 / 180.0);
    //get Theta_f in radians
    f2 = (theta + sc * f) * pi / 180.0;

    // Computing cos(Theta_f)^2 as in (24)
    cosf = cos(f2);
    cosf2 = cosf * cosf;
    p4 = -(pc) * p1 * (1 - cosf2) / (cosf2 + 1);

    return 0;
}

// anisotropic scattering
int anisoscat(double& nxp, double& nyp, double& nzp, double& sint, double& cost, double& sinp, double& cosp, double& phi, double& fi, double& fq, double& fu, double& fv, double& pl, double& pc, double& sc, double& hgg, double& g2) {

    double tmp,a, b, p1, p2, p3, p4, a11, a12, a13, a21, a22, a23, a24, a31, a32, a33, a34, a42, a43, a44, si, sq, su, sv, ri1, ri3, cosb, cos2, sin2, phip, wght, bott, cosb2, cosi1, cosi2, cosi3, sini1, sini2, sini3, cos2i1, cos2i2, cos2i3, sin2i1, sin2i2, sin2i3, sinbt, costp, sintp, cosdph, sin2cos1, cos2sin1;

    double pi = M_PI;
    double twopi = 2 * M_PI;

    // Normalizing Stoke's parameters
    wght = fi;
    fi /= wght;
    fq /= wght;
    fu /= wght;
    fv /= wght;

    //Store the initial variables
    costp = cost;
    sintp = sint;
    phip = phi;

    // Computing cos(Theta) according to (33)
    tmp = (1.0 - g2) / (1.0 - hgg + hgg * 2.0 * urand());
    cosb = (g2 + 1.0 - tmp * tmp) / (hgg * 2.0);

    // Computing cosb2=cos(Theta)^2 and b=sin(Theta)^2
    tmp = cosb;
    cosb2 = tmp * tmp;
    b = cosb2 - 1.0;

    // Run dustmat to calculate p1...p4 of the dust matrix and sample a new Theta
    dustmat(p1, p2, p3, p4, cosb, cosb2, pl, pc, sc, hgg, g2, pi);


    a = p1;

    // Correct the mistakes (if someting bad happened to the cos(Theta)^2, cos(Theta)  and sin(Theta)^2)
    if (fabs(cosb) > 1.0) {
        if (cosb > 1.0) {
            cosb = 1.0;
            cosb2 = 1.0;
            b = 0.0;
        } else {
            cosb = -1.0;
            cosb2 = 1.0;
            b = 0.0;
        }
    }
    //Compute sin(Theta) properly
    sinbt = sqrt(1.0 - cosb2);

    // Sample a random i1 (ri1) from 2*pi. Use it to calculate the new set of Stokes parameters from (27).
    ri1 = twopi * urand();

    if (ri1 > pi) {
        //Reduce ri1 to ri3 which can take values from 0 to pi
        ri3 = twopi - ri1;

        //Calculate the elements of the rotation matrix L(psi), where psi=ri3
        cosi3 = cos(ri3);
        sini3 = sin(ri3);
        sin2i3 = sini3 * 2.0 * cosi3;
        cos2i3 = cosi3 * 2.0 * cosi3 - 1.0;

        //If the scattering angle is 0, return, if it is pi just invert the U
        if (cosb == 1.0) {
            return 0;
        } else {
            if (cosb == -1.0) {
                fu = -(fu);
                return 0;
            }
        }

        //Calculate the new scattering cosine, sine and the relevant measures of the i2
        cost = costp * cosb + sintp * sinbt * cosi3;
        if (fabs(cost) < 1.0) {
            sint = (tmp = sqrt(1.0 - cost * cost), fabs(tmp));
            sini2 = sini3 * sintp / sint;
            bott = sint * sinbt;
            cosi2 = costp / bott - cost * cosb / bott;
        } else {
            sint = 0.0;
            sini2 = 0.0;
            if (cost >= 1.0) {
                cosi2 = -1.0;
            }
            if (cost <= -1.0) {
                cosi2 = 1.0;
            }
        }

        //calculate the difference of the scattering angle phi, and the new angle
        cosdph = -cosi2 * cosi3 + sini2 * sini3 * cosb;

        if (fabs(cosdph) > 1.0) {
            if (cosdph > 1.0) {
                cosdph = 1.0;
            } else {
                cosdph = -1.0;
            }
        }

        phi = phip + acos(cosdph);

        //make sure phi is from 0 to two pi
        if (phi > twopi) {
            phi -= twopi;
        }
        if (phi < 0.0) {
            phi += twopi;
        }

        //calculate the elements for the L(pi - i2)
        sin2i2 = sini2 * 2.0 * cosi2;
        cos2i2 = cosi2 * 2.0 * cosi2 - 1.0;
        sin2 = sin2i2 * sin2i3;
        cos2 = cos2i2 * cos2i3;
        sin2cos1 = sin2i2 * cos2i3;
        cos2sin1 = cos2i2 * sin2i3;

        //finally, calculate the final elements of the matrix S (27)
        a11 = p1;
        a12 = p2 * cos2i3;
        a13 = p2 * sin2i3;
        a21 = p2 * cos2i2;
        a22 = p1 * cos2 - p3 * sin2;
        a23 = p1 * cos2sin1 + p3 * sin2cos1;
        a24 = -p4 * sin2i2;
        a31 = -p2 * sin2i2;
        a32 = -p1 * sin2cos1 - p3 * cos2sin1;
        a33 = -p1 * sin2 + p3 * cos2;
        a34 = -p4 * cos2i2;
        a42 = -p4 * sin2i3;
        a43 = p4 * cos2i3;
        a44 = p3;

        //if random i1 < pi
    } else {
        //Calculate the elements of the rotation matrix L(psi), where psi=ri1
        cosi1 = cos(ri1);
        sini1 = sin(ri1);
        sin2i1 = sini1 * 2.0 * cosi1;
        cos2i1 = cosi1 * 2.0 * cosi1 - 1.0;

        //If the scattering angle is 0, return, if it is pi just invert the U
        if (cosb == 1.0) {
            return 0;
        } else {
            if (cosb == -1.0) {
                fu = -(fu);
                return 0;
            }
        }

        //Calculate the new scattering cosine, sine and the relevant measures of the i2
        cost = costp * cosb + sintp * sinbt * cosi1;
        if (fabs(cost) < 1.0) {
            sint = (tmp = sqrt(1.0 - cost * cost), fabs(tmp));
            sini2 = sini1 * sintp / sint;
            bott = sint * sinbt;
            cosi2 = costp / bott - cost * cosb / bott;
        } else {
            sint = 0.0;
            sini2 = 0.0;
            if (cost >= 1.0) {
                cosi2 = -1.0;
            }
            if (cost <= -1.0) {
                cosi2 = 1.0;
            }
        }

        //calculate the difference of the scattering angle phi, and the new angle
        cosdph = -cosi1 * cosi2 + sini1 * sini2 * cosb;
        if (fabs(cosdph) > 1.0) {
            if (cosdph > 1.0) {
                cosdph = 1.0;
            } else {
                cosdph = -1.0;
            }
        }

        phi = phip - acos(cosdph);

        //make sure phi is from 0 to two pi
        if (phi > twopi) {
            phi -= twopi;
        }
        if (phi < 0.0) {
            phi += twopi;
        }

        //calculate the elements for the L(pi - i2)
        sin2i2 = sini2 * 2.0 * cosi2;
        cos2i2 = cosi2 * 2.0 * cosi2 - 1.0;
        sin2 = sin2i2 * sin2i1;
        cos2 = cos2i2 * cos2i1;
        sin2cos1 = sin2i2 * cos2i1;
        cos2sin1 = cos2i2 * sin2i1;

        //finally, calculate the final elements of the matrix S (27)
        a11 = p1;
        a12 = p2 * cos2i1;
        a13 = -p2 * sin2i1;
        a21 = p2 * cos2i2;
        a22 = p1 * cos2 - p3 * sin2;
        a23 = -p1 * cos2sin1 - p3 * sin2cos1;
        a24 = p4 * sin2i2;
        a31 = p2 * sin2i2;
        a32 = p1 * sin2cos1 + p3 * cos2sin1;
        a33 = -p1 * sin2 + p3 * cos2;
        a34 = -p4 * cos2i2;
        a42 = p4 * sin2i1;
        a43 = p4 * cos2i1;
        a44 = p3;
    }

    //calculate S * [I,Q,U,V]
    si = (a11 * fi + a12 * fq + a13 * fu) / a;
    sq = (a21 * fi + a22 * fq + a23 * fu + a24 * fv) / a;
    su = (a31 * fi + a32 * fq + a33 * fu + a34 * fv) / a;
    sv = (a42 * fq + a43 * fu + a44 * fv) / a;

    // Compute final Stokes pars
    fi = si * wght;
    fq = sq * wght;
    fu = su * wght;
    fv = sv * wght;

    //Calculate the final cosines and the new photon coordinates
    cosp = cos(phi);
    sinp = sin(phi);
    nxp = sint * cosp;
    nyp = sint * sinp;
    nzp = cost;

    return 0;
}

int findTauMax(double xp, double yp, double zp, double nxp, double nyp, double nzp, double xmax, double ymax, double zmax, double* xface, double* yface, double* zface, double& tau1) {

    double r__1;
    static double d__, dx, dy, dz, dsx, dsy, dsz, smax, xcur, ycur, zcur, dcell;
    static int celli, cellj, cellk;
    static double taurun, taucell;



    // set the cumulative distance and optical depth (d and taurun)
    // along the photon path to zero.  set the current photon coordinates.
    // note that the origin of the (xcur,ycur,zcur) system is at the
    // bottom corner of the grid.


    taurun = 0.0;
    taucell = 0.0;
    d__ = 0.0;
    dcell = 0.0;
    xcur = xp + xmax;
    ycur = yp + ymax;
    zcur = zp + zmax;
    celli = xcell;
    cellj = ycell;
    cellk = zcell;
    // compute max distance to the edge of the grid, smax
    if (nxp > 0.0) {
        dsx = (xmax * 2.0 - xcur) / nxp;
    } else if (nxp < 0.0) {
        dsx = -xcur / nxp;
    } else if (nxp == 0.0) {
        dsx = xmax * 100.0;
    }
    if (nyp > 0.0) {
        dsy = (ymax * 2.0 - ycur) / nyp;
    } else if (nyp < 0.0) {
        dsy = -ycur / nyp;
    } else if (nyp == 0.0) {
        dsy = ymax * 100.0;
    }
    if (nzp > 0.0) {
        dsz = (zmax * 2.0 - zcur) / nzp;
    } else if (nzp < 0.0) {
        dsz = -zcur / nzp;
    } else if (nzp == 0.0) {
        dsz = zmax * 100.0;
    }

    r__1 = min(dsx,dsy);
    smax = min(r__1,dsz);
    if (smax < .001f) {
        tau1 = 0.0;
        return 0;
    }
    //integrate through grid
    while(d__ < smax * .999f) {
        //optical depth to next cell wall is
        // taucell= (distance to cell)*(opacity of current cell)
        //         	cout << "CELLi:"<<celli << " CELLj:"<<cellj << " CELLk:"<<cellk<<endl;
        taucell = dcell * rhokappa[celli][cellj][cellk];
        // if taurun+taucell<tau then photon moves distance dcell
        // (i.e. ends up on next cell wall) and update photon position
        // and cell.
        taurun += taucell;
        xcur += dcell * nxp;
        ycur += dcell * nyp;
        zcur += dcell * nzp;

        // calculate grid cell coordinates
        celli = (int) (xcur * xdim / (xmax * 2.0))-1;
        cellj = (int) (ycur * ydim / (ymax * 2.0))-1;
        cellk = (int) (zcur * zdim / (zmax * 2.0))-1;

        // find distance to next x, y, and z cell walls.
        // note that dx is not the x-distance, but the actual distance along
        // the direction of travel to the next x-face, and likewise for dy and dz.
        if (nxp > 0.0) {
            dx = (xface[celli + 1] - xcur) / nxp;
            if (dx < 1e-5f) {
                xcur = xface[celli + 1];
                ++celli;
                dx = (xface[celli + 1] - xcur) / nxp;
            }
        } else if (nxp < 0.0) {
            dx = (xface[celli] - xcur) / nxp;
            if (dx < 1e-5f) {
                xcur = xface[celli];
                dx = (xface[celli - 1] - xcur) / nxp;
                --celli;
            }
        } else if (nxp == 0.0) {
            dx = xmax * 100.0;
        }
        if (nyp > 0.0) {
            dy = (yface[cellj + 1] - ycur) / nyp;
            if (dy < 1e-5f) {
                ycur = yface[cellj + 1];
                ++cellj;
                dy = (yface[cellj + 1] - ycur) / nyp;
            }
        } else if (nyp < 0.0) {
            dy = (yface[cellj] - ycur) / nyp;
            if (dy < 1e-5f) {
                ycur = yface[cellj];
                dy = (yface[cellj - 1] - ycur) / nyp;
                --cellj;
            }
        } else if (nyp == 0.0) {
            dy = ymax * 100.0;
        }
        if (nzp > 0.0) {
            dz = (zface[cellk + 1] - zcur) / nzp;
            if (dz < 1e-5f) {
                zcur = zface[cellk + 1];
                ++cellk;
                dz = (zface[cellk + 1] - zcur) / nzp;
            }
        } else if (nzp < 0.0) {
            dz = (zface[cellk] - zcur) / nzp;
            if (dz < 1e-5f) {
                zcur = zface[cellk];
                dz = (zface[cellk - 1] - zcur) / nzp;
                --cellk;
            }
        } else if (nzp == 0.0) {
            dz = zmax * 100.0;
        }
        // distances are only zero if photon is on cell wall.  if it is
        // on cell wall then set to arbitrary large distance, since we will
        // in fact hit another wall
        if (dx == 0.0 || fabs(dx) < 1e-6f) {
            dx = xmax * 100.0;
        }
        if (dy == 0.0 || fabs(dy) < 1e-6f) {
            dy = ymax * 100.0;
        }
        if (dz == 0.0 || fabs(dz) < 1e-6f) {
            dz = zmax * 100.0;
        }
        //find distance to next cell wall
        r__1 = min(dx,dy);
        dcell = min(r__1,dz);
        if (dx < 0.0) {
            dcell = min(dy,dz);
        }
        if (dy < 0.0) {
            dcell = min(dx,dz);
        }
        if (dz < 0.0) {
            dcell = min(dx,dy);
        }
        d__ += dcell;
    }

    //finally, return the tau of the integration run
    tau1 = taurun;

    return 0;
}



int integration(double& xp, double& yp, double& zp, double nxp, double nyp, double nzp, double xmax, double ymax, double zmax, double* xface, double* yface, double* zface, int& tflag) {

    double r__1;
    static double d__, d1, dx, dy, dz, tau, dsx, dsy, dsz;
    static double smax, xcur, ycur, zcur, dcell;
    static int celli, cellj, cellk;
    static double taurun, taucell;


    tflag = 0;
    //generate random optical depth tau
    tau = -log(urand());
    // set the cumulative distance and optical depth (d and taurun)
    // along the photon path to zero.  set the current photon coordinates.
    // note that the origin of the (xcur,ycur,zcur) system is at the
    // bottom corner of the grid.
    taurun = 0.0;
    d__ = 0.0;
    xcur = xp + xmax;
    ycur = yp + ymax;
    zcur = zp + zmax;
    celli = xcell;
    cellj = ycell;
    cellk = zcell;
    // calculate maximum distance photon can travel
    if (nxp > 0.0) {
        dsx = (xmax * 2.0 - xcur) / nxp;
    } else if (nxp < 0.0) {
        dsx = -xcur / nxp;
    } else if (nxp == 0.0) {
        dsx = xmax * 100.0;
    }
    if (nyp > 0.0) {
        dsy = (ymax * 2.0 - ycur) / nyp;
    } else if (nyp < 0.0) {
        dsy = -ycur / nyp;
    } else if (nyp == 0.0) {
        dsy = ymax * 100.0;
    }
    if (nzp > 0.0) {
        dsz = (zmax * 2.0 - zcur) / nzp;
    } else if (nzp < 0.0) {
        dsz = -zcur / nzp;
    } else if (nzp == 0.0) {
        dsz = zmax * 100.0;
    }

    r__1 = min(dsx,dsy);
    smax = min(r__1,dsz);
    if (smax < .001) {
        tflag = 1;
        return 0;
    }
    // integrate through grid
    while(taurun < tau && d__ < smax * .999) {
        // find distance to next x, y, and z cell walls.
        // note that dx is not the x-distance, but the actual distance along
        // the direction of travel to the next x-face, and likewise for dy and dz.
        if (nxp > 0.0) {
            dx = (xface[celli + 1] - xcur) / nxp;
            if (dx < 1e-5f) {
                xcur = xface[celli + 1];
                ++celli;
                dx = (xface[celli + 1] - xcur) / nxp;
            }
        } else if (nxp < 0.0) {
            dx = (xface[celli] - xcur) / nxp;
            if (dx < 1e-5f) {
                xcur = xface[celli];
                dx = (xface[celli - 1] - xcur) / nxp;
                --celli;
            }
        } else if (nxp == 0.0) {
            dx = xmax * 100.0;
        }
        if (nyp > 0.0) {
            dy = (yface[cellj + 1] - ycur) / nyp;
            if (dy < 1e-5f) {
                ycur = yface[cellj + 1];
                ++cellj;
                dy = (yface[cellj + 1] - ycur) / nyp;
            }
        } else if (nyp < 0.0) {
            dy = (yface[cellj] - ycur) / nyp;
            if (dy < 1e-5f) {
                ycur = yface[cellj];
                dy = (yface[cellj - 1] - ycur) / nyp;
                --cellj;
            }
        } else if (nyp == 0.0) {
            dy = ymax * 100.0;
        }
        if (nzp > 0.0) {
            dz = (zface[cellk + 1] - zcur) / nzp;
            if (dz < 1e-5f) {
                zcur = zface[cellk + 1];
                ++cellk;
                dz = (zface[cellk + 1] - zcur) / nzp;
            }
        } else if (nzp < 0.0) {
            dz = (zface[cellk] - zcur) / nzp;
            if (dz < 1e-5) {
                zcur = zface[cellk];
                dz = (zface[cellk - 1] - zcur) / nzp;
                --cellk;
            }
        } else if (nzp == 0.0) {
            dz = zmax * 100.0;
        }
        // distances are only zero if photon is on cell wall.  if it is
        // on cell wall then set to arbitrary large distance, since we will
        // in fact hit another wall
        if (dx == 0.0 || fabs(dx) < .001) {
            dx = xmax * 100.0;
        }
        if (dy == 0.0 || fabs(dy) < .001) {
            dy = ymax * 100.0;
        }
        if (dz == 0.0 || fabs(dz) < .001) {
            dz = zmax * 100.0;
        }
        // find distance to next cell wall -- minimum of dx, dy, and dz
        r__1 = min(dx,dy);
        dcell = min(r__1,dz);
        if (dx < 0.0) {
            dcell = min(dy,dz);
        }
        if (dy < 0.0) {
            dcell = min(dx,dz);
        }
        if (dz < 0.0) {
            dcell = min(dx,dy);
        }
        // optical depth to next cell wall is
        // taucell= (distance to cell)*(opacity of current cell)
        //         	cout << "CELLi:"<< celli<< "CELLj:"<< cellj<< "CELLk:"<< cellk<<endl;
        taucell = dcell * rhokappa[celli] [cellj] [cellk];
	
        // if taurun+taucell>tau then scatter at distance d+d1.
        // update photon position and cell.
        // if taurun+taucell<tau then photon moves distance dcell
        // (i.e. ends up on next cell wall) and update photon position
        // and cell.
        if (taurun + taucell >= tau) {
            d1 = (tau - taurun) / rhokappa[celli][cellj][cellk];
            d__ += d1;
            taurun += taucell;
            xcur += d1 * nxp;
            ycur += d1 * nyp;
            zcur += d1 * nzp;
            // calculate grid coords (note that there is some added bias introduced,
            // because of attempts to avoid buffer overflow)
            celli = abs((int) (xcur * xdim / (xmax * 2.0))-1);
            cellj = abs((int) (ycur * ydim / (ymax * 2.0))-1);
            cellk = abs((int) (zcur * zdim / (zmax * 2.0))-1);
        } else {
            d__ += dcell;
            taurun += taucell;
            xcur += dcell * nxp;
            ycur += dcell * nyp;
            zcur += dcell * nzp;
            // calculate grid coords
            celli = abs((int) (xcur * xdim / (xmax * 2.0))-1) ;
            cellj = abs((int) (ycur * ydim / (ymax * 2.0))-1) ;
            cellk = abs((int) (zcur * zdim / (zmax * 2.0))-1);
        }
    }
    
    // calculate photon final position.  if it escapes envelope then
    // set tflag=1.  if photon doesn't escape leave tflag=0 and update
    // photon position.
    if (d__ >= smax * .999) {
        tflag = 1;
    } else {
        xp += d__ * nxp;
        yp += d__ * nyp;
        zp += d__ * nzp;
        xcell = abs((int) ((xp + xmax) * xdim / (xmax * 2.0))-1);
        ycell = abs((int) ((yp + ymax) * ydim / (ymax * 2.0))-1);
        zcell = abs((int) ((zp + zmax) * zdim / (zmax * 2.0))-1);
    }
    return 0;
}


#endif
